###
### Code for running the RJMCMC for geostatistical regression models
### Author: Devin S. Johnson <devin.johnson@noaa.gov>
### Date: 7/6/2011
###

## 
## HELPER FUNCTIONS 
## these are used in the RJMCMC main functions below
##
"sp.cov.exp" <- function(D, t1, t2, phi)
{
	exp(t1-D/phi) + diag(rep(exp(t2),dim(D)[1]))
}

"sp.cov.exp.nn" <- function(D, t1, phi)
{
	exp(t1-D/phi)
}


"bdist" <- function(mk, beta)
{
	ind <- !mk==0
	p <- diag(mk)[,ind]
	p %*% beta	 
}

"XColSelect" <- function(mk, X)
{
	ind <- !mk==0
	Xk <- as.matrix(X[,ind])
}

"dist.fun" <- function(x, y) {
	n1 <- length(x)
	ones <- rep(1,n1)
	sqrt( 
		( matrix(x,n1,1)%*%matrix(ones,1,n1)-matrix(ones,n1,1)%*%matrix(x,1,n1) )^2 +
		( matrix(y,n1,1)%*%matrix(ones,1,n1)-matrix(ones,n1,1)%*%matrix(y,1,n1) )^2 
	)
}

"log.GRF" <- function(z, mu.z, Sig.i)
{
	Q <- t(z - mu.z)%*%Sig.i%*%(z - mu.z)
	as.numeric(-(length(z)*log(2*pi) - determinant(Sig.i)$modulus + Q)/2 )
}

"log.tri" <- function(x)
{
	log(1-abs(x))
} 

"b.prior" <- function(z, mk, g)
{
	s <- sum(mk)
	if(s==1){as.matrix(var(z))}
	else{diag(c(var(z),g*rep(1,s-1)))}
}

"log.Pois" <- function(y, mu)
{
	sum(-mu + y*log(mu) - lfactorial(y))
}

"drift" <- function(y, z, S.i, mu.z)
{
	y - exp(z) - S.i%*%(z - mu.z)
}

"log.Pmk" <- function(mk,prob)
{
	mk[-1]%*%log(prob) + (1-mk[-1])%*%log(1-prob)
}
#############################################################################

##
## RJMCMC FOR GAUSSIAN DATA
#
#As used in example 1 of the paper
#spatial.rjmcmc(
#	whiptail, 
#	g=100, mu.Int=mean(whiptail[,3]), mu.theta=c(0, 0, 0, 0), v.theta=c(10, 10, 1, 1), 
#	model.probs=rep(0.5,7),
#	parm.start=as.vector(scan("D:\\spatial\\R_output\\lizard\\log_trans\\last.txt")),
#	burn=0, mh=100000,
#	theta.tune=c(1.3, 0.45, 2.4, 2.4, 2.0),
#	out.dir="D:\\spatial\\R_output\\lizard\\log_trans\\"
#)

"spatial.rjmcmc" <- function(dataset, g, mu.Int, mu.theta, v.theta, model.probs,
	parm.start, burn, mh, theta.tune, h, out.dir)
{
require(mvtnorm)
start <- Sys.time()
z <- as.matrix(dataset[,3])
X <- as.matrix(dataset[,-c(1:3)])
Intercept <- rep(1,length(z))
X <- cbind(Intercept, X)
K <- dim(X)[2]
C <- as.matrix(dataset[,1:2])
# MCMC parameters ##
iter <- burn + mh
model.out <- matrix(0,mh,K)
beta.out <- matrix(0,mh,K)
cov.out <- matrix(0,mh,5)
t1.accept <- rep(0,iter)
t2.accept <- rep(0,iter)
t3.accept <- rep(0,iter)
t4.accept <- rep(0,iter)
psi.accept <- rep(0,iter)
# Priors
mu.t1 <- mu.theta[1]
mu.t2 <- mu.theta[2]
mu.t3 <- mu.theta[3]
mu.t4 <- mu.theta[4]
v.t1 <- v.theta[1]
v.t2 <- v.theta[2]
v.t3 <- v.theta[3]
v.t4 <- v.theta[4]
# Set initial values ##
t1 <- parm.start[1]
t2 <- parm.start[2]
t3 <- parm.start[3]
t4 <- parm.start[4]
psi <- parm.start[5]
B <- matrix(c(exp(t3),psi*exp(0.5*(t3+t4)),psi*exp(0.5*(t3+t4)),exp(t4)),2,2)
D <- as.matrix(dist(C%*%t(chol(B))))
S <- sp.cov.exp(D,t1, t2, 1)
S.i <- solve(S)
mk <- parm.start[6:(6+K-1)]
Xk <- XColSelect(mk, X)
betak <- parm.start[(6+K-1):(6+2*K-1)]
betak <- parm.start[(3+K):(3+K+sum(mk)-1)]
mu.b.f <- c(mu.Int,rep(0,K-1))
mu.b <- mu.b.f[mk==1]
mu.z <- Xk%*%betak
Vk <- g*(exp(t1)+exp(t2)) * solve(t(Xk)%*%Xk)
Vk.i <- solve(Vk)
A <- Xk %*% Vk %*% t(Xk) + S
A.i <- solve(A)
tune.t1 <- theta.tune[1]
tune.t2 <- theta.tune[2]
tune.t3 <- theta.tune[3]
tune.t4 <- theta.tune[4]
tune.psi <- theta.tune[5]
for(i in 1:iter)
{
# Update theta1 = log(sig2)
	t1.new <- rnorm(1, t1, tune.t1)
	S.new <- sp.cov.exp(D, t1.new, t2, 1)
	S.new.i <- solve(S.new)
	R <- exp(
		log.GRF(z, mu.z, S.new.i) - log.GRF(z, mu.z, S.i) + 
		dnorm(t1.new,mu.t1,sqrt(v.t1), log=T) - dnorm(t1,mu.t1,sqrt(v.t1), log=T)
		)
	if(runif(1)<R){
		t1 <- t1.new
		S <- S.new
		S.i <- S.new.i
		t1.accept[i] <- 1
	}
	if(iter>burn){ cov.out[(i-burn),1] <- t1 }	
# Update theta2 = log(tau2)
	t2.new <- rnorm(1, t2, tune.t2)
	S.new <- sp.cov.exp(D, t1, t2.new, 1)
	S.new.i <- solve(S.new)
	R <- exp(
		log.GRF(z, mu.z, S.new.i) - log.GRF(z, mu.z, S.i) + 
		dnorm(t2.new,mu.t2,sqrt(v.t2), log=T) - dnorm(t2,mu.t2,sqrt(v.t2), log=T)
		)
	if(runif(1)<R){
		t2 <- t2.new
		S <- S.new
		S.i <- S.new.i
		t2.accept[i] <- 1
	}
	if(iter>burn){ cov.out[(i-burn),2] <- t2 }
# Update theta3 = log(aniso1)
	t3.new <- rnorm(1, t3, tune.t3)
	B.new <- matrix(c(exp(t3.new),psi*exp(0.5*(t3.new+t4)),psi*exp(0.5*(t3.new+t4)),exp(t4)),2,2)
	D.new <- as.matrix(dist(C%*%t(chol(B.new))))
	S.new <- sp.cov.exp(D.new, t1, t2, 1)
	S.new.i <- solve(S.new)
	R <- exp(
		log.GRF(z, mu.z, S.new.i) - log.GRF(z, mu.z, S.i) + 
		dnorm(t3.new,mu.t3,sqrt(v.t3), log=T) - dnorm(t3,mu.t3,sqrt(v.t3), log=T)
		)
	if(runif(1)<R){
		t3 <- t3.new
		S <- S.new
		S.i <- S.new.i
		B <- B.new
		D <- D.new
		t3.accept[i] <- 1
	}
	if(iter>burn){ cov.out[(i-burn),3] <- t3 }
# Update theta4 = log(aniso2)
	t4.new <- rnorm(1, t4, tune.t4)
	B.new <- matrix(c(exp(t3),psi*exp(0.5*(t3+t4.new)),psi*exp(0.5*(t3+t4.new)),exp(t4.new)),2,2)
	D.new <- as.matrix(dist(C%*%t(chol(B.new))))
	S.new <- sp.cov.exp(D.new, t1, t2, 1)
	S.new.i <- solve(S.new)
	R <- exp(
		log.GRF(z, mu.z, S.new.i) - log.GRF(z, mu.z, S.i) + 
		dnorm(t4.new,mu.t4,sqrt(v.t4), log=T) - dnorm(t4,mu.t4,sqrt(v.t4), log=T)
		)
	if(runif(1)<R){
		t4 <- t4.new
		S <- S.new
		S.i <- S.new.i
		B <- B.new
		D <- D.new
		t4.accept[i] <- 1
	}
	if(iter>burn){ cov.out[(i-burn),4] <- t4 }
# Update psi
	up <- min(1, psi+tune.psi)
	lo <- max(-1, psi-tune.psi)
	psi.new <-  runif(1,lo,up) 
	up.new <- min(1, psi.new+tune.psi)
	lo.new <- max(-1, psi.new-tune.psi)
	B.new <- matrix(c(exp(t3),psi.new*exp(0.5*(t3+t4)),psi.new*exp(0.5*(t3+t4)),exp(t4)),2,2)
	D.new <- as.matrix(dist(C%*%t(chol(B.new))))
	S.new <- sp.cov.exp(D.new, t1, t2, 1)
	S.new.i <- solve(S.new)
	R <- exp(
		log.GRF(z, mu.z, S.new.i) - log.GRF(z, mu.z, S.i) + 
		log.tri(psi.new) - log.tri(psi) +
		dunif(psi,lo.new,up.new,log=T) - dunif(psi.new,lo,up,log=T)
	)
	if(runif(1)<R){
		psi <- psi.new
		B <- B.new
		D <- D.new
		S <- S.new
		S.i <- S.new.i
		psi.accept[i] <- 1
	}
	if(iter>burn){ cov.out[(i-burn),5] <- psi }	
# Model update
	mk.new <- mk
	prop <- sample(2:K, 1)
	mk.new[prop] <- ifelse(mk.new[prop]==0,1,0)
	Xk.new <- XColSelect(mk.new, X)
     	Vk.new <- g*exp(t1)*solve(t(Xk.new)%*%Xk.new)
	mu.b.new <- mu.b.f[mk.new==1]
     	A.new <- Xk.new %*% Vk.new %*% t(Xk.new) + S;
	A.new.i <- solve(A.new)
     	R <- exp(
		log.GRF(z, Xk.new%*%mu.b.new, A.new.i) - log.GRF(z, Xk%*%mu.b, A.i) +
		log.Pmk(mk.new,model.probs) - log.Pmk(mk,model.probs)
		)
     	if(runif(1) <= R){
        	mk <- mk.new
		mu.b <- mu.b.new
         	A <- A.new
		A.i <- A.new.i
         	Xk = Xk.new
         	Vk = Vk.new
		Vk.i <- solve(Vk.new)
     	}    
	if(iter>burn){ model.out[(i-burn),] <- mk }
# Update betas
	sig.beta <- solve(Vk.i + t(Xk) %*% S.i %*% Xk)
	betahat <- sig.beta %*% (Vk.i%*%mu.b + t(Xk)%*%S.i%*%z) 
	betak <- t(rmvnorm(1,betahat,sig.beta))
	mu.z <- Xk%*%betak
	if(iter>burn){ beta.out[(i-burn),] <- t(bdist(mk, betak)) }
} # End of MH loop
last <- as.vector(c(cov.out[mh,], mk, betak))
dimnames(cov.out)[[2]] <- c("Theta1", "Theta2", "Theta3", "Theta4", "Psi")
dimnames(model.out)[[2]] <- dimnames(X)[[2]]
dimnames(beta.out)[[2]] <- dimnames(X)[[2]]
write.table(beta.out, paste(out.dir,'betaout.txt',sep=''), sep='\t', quote=F, row.names=F)
write.table(cov.out, paste(out.dir,'covout.txt',sep=''), sep='\t', quote=F, row.names=F)
write.table(model.out, paste(out.dir,'modelout.txt',sep=''), sep='\t', quote=F, row.names=F)
write.table(last, paste(out.dir,'last.txt',sep=''), quote=F, row.names=F, col.names=F)
end <- Sys.time()
et <- difftime(end, start)
print(et)
print(paste("Theta 1 parameter acceptance rate: ", mean(t1.accept[(burn+1):iter])))
print(paste("Theta 2 parameter acceptance rate: ", mean(t2.accept[(burn+1):iter])))
print(paste("Theta 3 parameter acceptance rate: ", mean(t3.accept[(burn+1):iter])))
print(paste("Theta 4 parameter acceptance rate: ", mean(t4.accept[(burn+1):iter])))
print(paste("Psi parameter acceptance rate: ", mean(psi.accept[(burn+1):iter])))
}# End of function
####################################################################################

##
## RJMCMC FOR POISSON DATA
#
#As used in example 2 of the paper
#Pois.rjmcmc(
#  formula = as.formula(full.model),
#  coords=c('LON','LAT'),
#	data = fish,
#	g=100, mu.Int=log(mean(fish[,y])), mu.theta=c(0, 0), v.theta=c(10, 1),
#	model.probs=rep(0.5,9),
#	parm.start=scan('last_pois.txt'),
#	#parm.start=startVal,
#	burn=0, mh=305000,
#	theta.tune=c( 0.161007704421001, 1.09477906703511),
#   h= 0.101465233498417,
#   adapt=5000,
#	out.dir=getwd(),
#	report = 1000,
#	append=F,
#	model.update=T
#)
"Pois.rjmcmc" <-
function(
   formula, coords=c('x','y'), data, g, mu.Int, mu.theta, v.theta, model.probs,
	parm.start, burn, mh, theta.tune, h, adapt=5000, out.dir, acf=FALSE,
   GCdist='km', report=TRUE, append=TRUE, model.update=TRUE, derived=NULL)
{
require(mvtnorm)
require(fields)
require(geoR)
if(!missing(GCdist))
{
  if(GCdist=='mi') Rad <- 3963.34
  else if(GCdist=='nm') Rad <- 3438.461
  else Rad <- 6378.388
}
start <- Sys.time()
mf <- model.frame(formula,data)
y <- mf[,attr(attr(mf,"terms"),"response")]
X <- model.matrix(formula, data)
K <- dim(X)[2]
N <- dim(X)[1]
if(missing(GCdist)) D <- as.matrix(dist(data[,coords]))
else D <- rdist(earth2euclid(data[,coords], R=Rad))
#else D <- rdist.earth(data[,coords], R=Rad)
diag(D) <- 0
# MCMC parameters ##
iter <- burn + mh
update.num <- 1
if(is.numeric(model.update)) update.num <- model.update; model.update <- TRUE
model.out <- matrix(0,mh,K)
beta.out <- matrix(0,mh,K)
cov.out <- matrix(0,mh,2)
theta1.accept <- rep(0,iter)
theta3.accept <- rep(0,iter)
z.accept <- rep(0,iter)
# Priors
mu.t1 <- mu.theta[1]
mu.t3 <- mu.theta[2]
v.t1 <- v.theta[1]
v.t3 <- v.theta[2]
# Set initial values ##
t1 <- parm.start[1]
t3 <- parm.start[2]
S <- cov.spatial(D, cov.pars=c(1,exp(t3)),kappa=0.5)
ln.det <- 2*sum(log(diag(chol(S))))
mk <- parm.start[3:(3+K-1)]
Xk <- XColSelect(mk, X)
betak <- parm.start[(3+K):(3+K+sum(mk)-1)]
mu.b.f <- c(mu.Int,rep(0,K-1))
mu.b <- mu.b.f[mk==1]
lambda <- Xk%*%betak
z <- parm.start[(3+K+sum(mk)):(3+K+sum(mk)+N-1)]
z.orth <- solve(S, z-lambda)
drift <- y-exp(z)-exp(-2*t1)*z.orth

if(!missing(theta.tune))
{
  tune.t1 <- theta.tune[1]
  tune.t3 <- theta.tune[2]
}
else
{
  tune.t1 <- sqrt(v.t1)
  tune.t3 <- sqrt(v.t3)
}
mu.y <- mu.y2 <- rep(0,length(lambda))
if(is.numeric(report)) report.num <- report; report <- TRUE
for(i in 1:iter)
{
tune.adj <- (i>100 & i<adapt)
#Z update
	z.new <- z + rnorm(N, (h^2/2)*drift, h)
	z.orth.new <- solve(S, z.new-lambda)
	drift.rev <- y-exp(z.new)-exp(-2*t1)*z.orth.new
	R <-  min(1, exp(
		      log.Pois(y, exp(z.new)) - log.Pois(y, exp(z)) +
		      crossprod(-(z.new-lambda)/exp(2*t1),z.orth.new)/2 -
              crossprod(-(z-lambda)/exp(2*t1),z.orth)/2 +
		      sum(dnorm(z, z.new+(h^2/2)*drift.rev, h, log=T)) -
		          sum(dnorm(z.new, z+(h^2/2)*drift, h, log=T))
		    ))
	if(runif(1) < R)
  {
		z <- z.new
		drift <- drift.rev
		z.orth <- z.orth.new
		z.accept[i] <- 1
	}
  if(tune.adj) h <- min(max(1.0e-12,h + 10*(R-0.574)/i), 1.0e7)
# Update theta1 (t1=ln(sig))
	t1.new <- rnorm(1, t1, tune.t1)
	R <- min(1, exp(
  crossprod(-(z-lambda)*(exp(-2*t1.new)-exp(-2*t1)),z.orth)/2 - N*(t1.new-t1) +
		dnorm(t1.new,mu.t1,sqrt(v.t1), log=T) - dnorm(t1,mu.t1,sqrt(v.t1), log=T)
		))
	if(runif(1)<R)
  {
		t1 <- t1.new
		theta1.accept[i] <- 1
	}
	if(iter>burn){ cov.out[(i-burn),1] <- t1 }
	if(tune.adj) tune.t1 <- min(max(1.0e-8,tune.t1 + 10*(R-0.44)/i), 1.0e7)    
# Update t3 ( log(phi) )
	t3.new <- rnorm(1, t3, tune.t3)
	S.new <- cov.spatial(D, cov.pars=c(1,exp(t3.new)),kappa=0.5)
	ln.det.new <- 2*sum(log(diag(chol(S.new))))
	z.orth.new <- solve(S.new, z-lambda)
  R <- min(1, exp(
         (-crossprod((z-lambda)/exp(2*t1),z.orth.new) - ln.det.new)/2 -
		         (-crossprod((z-lambda)/exp(2*t1),z.orth) - ln.det)/2 +
         dnorm(t3.new,mu.t3,sqrt(v.t3), log=T) -
             dnorm(t3,mu.t3,sqrt(v.t3), log=T)
		))
	if(runif(1)<R)
  {
		t3 <- t3.new
		S <- S.new
		z.orth <- z.orth.new
		ln.det <- ln.det.new
		theta3.accept[i] <- 1
	}
	if(iter>burn){ cov.out[(i-burn),2] <- t3 }
  if(tune.adj) tune.t3 <- min(max(1.0e-8,tune.t3 + 10*(R-0.44)/i), 1.0e7)
# Model update
  if(model.update & i%%update.num==0)
  {
  	mk.new <- mk
  	prop <- sample(2:K, 1)
  	mk.new[prop] <- 1 - mk.new[prop]
  	Xk.new <- XColSelect(mk.new, X)
	  Vk <- g * exp(2*t1) * solve(crossprod(Xk))
  	Vk.new <- g*exp(2*t1)*solve(crossprod(Xk.new))
  	iVk.new <- exp(-2*t1)*crossprod(Xk.new)/g
  	mu.b.new <- mu.b.f[mk.new==1]
  	A <- Xk %*% Vk %*% t(Xk) + S*exp(2*t1)
    iAzmu <- solve(A, (z-Xk%*%mu.b))
    ln.detA <- 2*sum(log(diag(chol(A))))
    A.new <- Xk.new %*% Vk.new %*% t(Xk.new) + S*exp(2*t1)
    iAzmu.new <- solve(A.new, (z-Xk.new%*%mu.b.new))
    ln.detA.new <- 2*sum(log(diag(chol(A.new))))
  	R <- exp(
  		-(crossprod(z-Xk.new%*%mu.b.new, iAzmu.new) + ln.detA.new)/2 +
         (crossprod(z-Xk%*%mu.b, iAzmu) + ln.detA)/2 +
      log.Pmk(mk.new,model.probs) - log.Pmk(mk,model.probs)
  		)
   	if(runif(1) <= R)
    {
   	  mk <- mk.new
  		mu.b <- mu.b.new
   	  Xk <- Xk.new
   	}
	 if(iter>burn){ model.out[(i-burn),] <- mk }
  }
# Update betas
  iVk <- exp(-2*t1) * crossprod(Xk) / g
  iSz <- solve(S,z)
  iSXk <- solve(S,Xk)
	Vkhat <- solve(iVk + crossprod(Xk, exp(-2*t1)*iSXk))
	mukhat <- Vkhat %*% (iVk%*%mu.b + exp(-2*t1)*crossprod(Xk, iSz))
	betak <- t(rmvnorm(1,mukhat,Vkhat))
	lambda <- Xk%*%betak
  z.orth <- iSz-iSXk%*%betak
	if(iter>burn){ beta.out[(i-burn),] <- t(bdist(mk, betak)) }
# fitted values
mu.y <- ((i-1)*mu.y + exp(z))/i
mu.y2 <- ((i-1)*mu.y2 + exp(2*z))/i	
if(report & i%%report.num==0) cat(paste(i," iterations completed\n"))
} # End of MH loop
dimnames(cov.out)[[2]] <- c("Theta1", "Theta3")
dimnames(model.out)[[2]] <- dimnames(X)[[2]]
dimnames(beta.out)[[2]] <- dimnames(X)[[2]]
wd <- getwd()
setwd(out.dir)
colnm <- !append
write.table(beta.out, 'betaout.txt', sep='\t', quote=F, row.names=F, col.names=colnm, append=append)
write.table(cov.out, 'covout.txt', sep='\t', quote=F, row.names=F, col.names=colnm, append=append)
write.table(model.out, 'modelout.txt', sep='\t', quote=F, row.names=F, col.names=colnm, append=append)
write.table(data.frame(fitted.y=mu.y, sd.fitted.y=sqrt(mu.y2 - mu.y^2)), 'fitted.txt', sep='\t', quote=F, row.names=F, col.names=colnm, append=append)
write.table(c(t1,t3,mk,betak,z),'last_pois.txt', sep='\t', quote=F, row.names=F, col.names=F)
setwd(wd)

if(acf){
	acf.mat <- matrix(0,41,N)
	for(i in 1:N){acf.mat[,i] <- acf(z.out[,i],plot=F,lag.max=40)$acf}
	write.table(acf.mat, paste(out.dir,'acfmat.txt',sep=''), sep='\t', quote=F, row.names=F)
}
end <- Sys.time()
et <- difftime(end, start)
print(et)
#cat(paste("Theta1 acceptance rate: ", mean(theta1.accept[(burn+1):iter])))
#cat(paste("Theta3 acceptance rate: ", mean(theta3.accept[(burn+1):iter])))
#cat(paste("Z acceptance rate: ", mean(z.accept[(burn+1):iter])))
cat("Acceptance rates in last 100 iterations:\n")
cat(paste("Theta1: ", mean(theta1.accept[(iter-100+1):iter]),"\n",collapse=''))
cat(paste("Theta3: ", mean(theta3.accept[(iter-100+1):iter]),"\n",collapse=''))
cat(paste("Z: ", mean(z.accept[(iter-100+1):iter]),"\n",collapse=''))
cat('\n')
cat('Final value of tuning parameters:\n')
cat(paste('Theta1 tuner: ', tune.t1,'\n',collapse=''))
cat(paste('Theta3 tuner: ', tune.t3,'\n',collapse=''))
cat(paste('h: ', h,'\n',collapse=''))
cat('\n')
}# End of function

